#!/usr/bin/env python
import asyncio
import json
import logging
from ssc2ce import Deribit

conn = Deribit()

pending = {}

logging.basicConfig(format='%(asctime)s %(name)s %(funcName)s %(levelname)s %(message)s', level=logging.INFO)
logger = logging.getLogger("deribit-writer")

instruments = []


async def handle_instruments(data: dict):
    global instruments
    request_id = data["id"]
    del pending[request_id]
    print(json.dumps(data))
    instruments += list(set(
        [x["instrument_name"][:11] if not x["instrument_name"].endswith("-PERPETUAL") else x["instrument_name"] for x in
         data["result"]]))

    if not pending:
        print(instruments)
        await subscribe_books()


async def handle_book_summary_by_currency(data: dict):
    request_id = data["id"]
    del pending[request_id]
    print(json.dumps(data))


async def handle_currencies(data: dict):
    for currency in data["result"]:
        symbol = currency["currency"]
        print(json.dumps(data))
        request_id = await conn.get_instruments(symbol, callback=handle_instruments)
        pending[request_id] = symbol
        request_id = await conn.send_public(
            request={"method": "public/get_book_summary_by_currency", "params": {"currency": symbol}},
            callback=handle_instruments)
        pending[request_id] = symbol


async def get_currencies():
    await conn.get_currencies(handle_currencies)


async def subscribe_books():
    global instruments
    await conn.send_public(request={
        "method": "public/subscribe",
        "params": {
            "channels": [f"book.{i}.raw" for i in instruments]
        }
    })

    await conn.send_public(request={
        "method": "public/set_heartbeat",
        "params": {
            "interval": 15
        }
    })


output = open("deribit_dump.jsonl", "w")


def dump(msg: str):
    output.write(msg)
    output.write('\n')


def stop():
    asyncio.ensure_future(conn.ws.close())


conn.on_before_handling = dump
conn.on_connect_ws = get_currencies

loop = asyncio.get_event_loop()
loop.call_later(3600, stop)

try:
    loop.run_until_complete(conn.run_receiver())
except KeyboardInterrupt:
    print("Application closed by KeyboardInterrupt.")



from ssc2ce.coinbase_pro.auth import make_auth_params
from ssc2ce.common.session import SessionWrapper
from aiohttp import ClientResponseError


class Coinbase(SessionWrapper):
    """
    Handlers:
     - on_connect_ws - Called after the connection is established.
            If auth_type is not equivalent to AuthType.NONE,
            on_connect_ws will be set to self.auth_login;
     - on_close_ws - Called after disconnection, default value is None;
     - on_authenticated - Called after authentication is confirmed, default value is None;
     - on_token - Called after receiving a response to an authentication request, default value is None;
     - on_message - Called when a message is received, default value is self.handle_message;
     - on_handle_response - Called when the message from the exchange does not contain the request id;
     - on_response_error - Called when the response contains an error message.
    """

    def __init__(self,
                 auth_param: dict = None,
                 sandbox: bool = True):

        super().__init__()

        self.ws_api = "wss://ws-feed-public.sandbox.pro.coinbase.com" if sandbox \
            else "wss://ws-feed.pro.coinbase.com"

        self.rest_api = "https://api-public.sandbox.pro.coinbase.com" if sandbox \
            else "https://api.pro.coinbase.com"
        self.sandbox = sandbox
        self.auth_param = auth_param

    async def get_currencies(self):
        try:
            return await self.public_get("/currencies")
        except ClientResponseError as e:
            self.logger.error(e.message)
            return []

    async def get_products(self):
        try:
            return await self.public_get("/products")
        except ClientResponseError as e:
            self.logger.error(e.message)
            return []

    def _auth_headers(self, path, method, body=''):
        auth_params = make_auth_params(key=self.auth_param["api_key"],
                                       secret=self.auth_param["secret_key"],
                                       passphrase=self.auth_param["passphrase"],
                                       request_method=method,
                                       request_path_url=path,
                                       request_body=body)

        return {
            'Content-Type': 'application/json',
            'CB-ACCESS-SIGN': auth_params['signature'].decode('ascii'),
            'CB-ACCESS-TIMESTAMP': auth_params['timestamp'],
            'CB-ACCESS-KEY': auth_params['key'],
            'CB-ACCESS-PASSPHRASE': auth_params['passphrase'],
        }

    async def private_get(self, request_path):
        headers = self._auth_headers(path=request_path, method='GET', body='')

        async with self._session.get(self.rest_api + request_path, headers=headers) as response:
            if response.status == 200:
                return await response.json()
            else:
                self.logger.error(self.rest_api + request_path)
                response.raise_for_status()



import base64
import hashlib
import hmac
import time


def make_auth_params(key,
                     secret,
                     passphrase,
                     request_method: str,
                     request_path_url: str,
                     request_body: str):

    timestamp = str(int(time.time()))
    message = timestamp + request_method + request_path_url
    if request_body:
        message += request_body

    hmac_key = base64.b64decode(secret)
    message = message.encode('utf-8')
    signature = hmac.new(hmac_key, message, hashlib.sha256)
    signature_b64 = base64.b64encode(signature.digest())

    return dict(key=key,
                passphrase=passphrase,
                timestamp=timestamp,
                signature=signature_b64)


def make_auth_header(key,
                     secret,
                     passphrase,
                     request_method: str,
                     request_path_url: str,
                     request_body: str):

    auth_params = make_auth_params(key=key,
                                   secret=secret,
                                   passphrase=passphrase,
                                   request_method=request_method,
                                   request_path_url=request_path_url,
                                   request_body=request_body)

    return {
        'Content-Type': 'application/json',
        'CB-ACCESS-SIGN': auth_params['signature'].decode('ascii'),
        'CB-ACCESS-TIMESTAMP': auth_params['timestamp'],
        'CB-ACCESS-KEY': auth_params['key'],
        'CB-ACCESS-PASSPHRASE': auth_params['passphrase'],
    }



#!/usr/bin/env python
import asyncio
import logging

from ssc2ce import Bitfinex

logging.basicConfig(format='%(asctime)s %(name)s %(funcName)s %(levelname)s %(message)s', level=logging.INFO)
logger = logging.getLogger("bitfinex-basic-example")

conn = Bitfinex()


def handle_subscription(data, self):
    print(data, self.receipt_time * 1000 - data[-1])


async def subscribe():
    await conn.subscribe({
        "channel": "ticker",
        "symbol": "tBTCUSD"
    }, handler=handle_subscription)


conn.on_connect_ws = subscribe


def stop():
    asyncio.ensure_future(conn.stop())


loop = asyncio.get_event_loop()
loop.call_later(3600, stop)

try:
    loop.run_until_complete(conn.run_receiver())
except KeyboardInterrupt:
    print("Application closed by KeyboardInterrupt.")



import re
from typing import Pattern


def resolve_route(value, routes):
    key, handler = None, None
    for key, handler in routes:
        if key:
            if isinstance(key, str):
                if key == value:
                    return handler
            elif isinstance(key, Pattern) and re.match(key, value):
                return handler

    if key is not None and key == "" and handler:
        return handler


class IntId:
    def __init__(self):
        self.id = 0

    def get_id(self):
        self.id += 1
        return self.id



#!/usr/bin/env python
import asyncio
import logging
import os

import aiohttp
from ssc2ce import Cex
from dotenv import load_dotenv

logging.basicConfig(format='%(asctime)s %(name)s %(funcName)s %(levelname)s %(message)s', level=logging.INFO)
logger = logging.getLogger("bitfinex-basic-example")

env_path = os.path.join(os.path.dirname(__file__), '.env')
load_dotenv(env_path)

auth_param = dict(apiKey=os.environ.get('CEX_API_KEY'),
                  secret=os.environ.get('CEX_SECRET'),
                  uid=os.environ.get('CEX_UID'))

conn = Cex(auth_param)

archive_orders = []


async def get_archive_orders(symbol1, symbol2):
    global archive_orders
    res = await conn.private_get(f"archived_orders/{symbol1}/{symbol2}")
    archive_orders += res


async def start():
    request = {'e': "get-balance", 'oid': '1'}

    await conn.ws.send_json(request)

    my_balance = {}
    data = await conn.ws.receive()
    if data.type == aiohttp.WSMsgType.TEXT:
        data = data.json()
        if data.get('ok') == 'ok':
            balance = data['data']['balance']
            for key, value in balance.items():
                if float(value) != 0.00:
                    my_balance[key] = value

    print(my_balance)

    coroutines = []
    data = await conn.public_get("currency_limits")

    if data.get('ok') == 'ok':
        for pair in data['data']['pairs']:
            if pair['symbol1'] in my_balance and pair['symbol2'] in my_balance:
                print(pair)
                coroutines.append(get_archive_orders(pair['symbol1'], pair['symbol2']))
                break

        if coroutines:
            await asyncio.gather(*coroutines)
            for order in archive_orders:
                print(order)

    await conn.stop()


conn.on_connect_ws = start
loop = asyncio.get_event_loop()

try:
    loop.run_until_complete(conn.run_receiver())
except KeyboardInterrupt:
    logger.info("Application closed by KeyboardInterrupt.")



#!/usr/bin/env python
import asyncio
import logging
import os

from examples.common.book_watcher import BookWatcher
from ssc2ce import Cex, CexParser
from dotenv import load_dotenv

logging.basicConfig(format='%(asctime)s %(name)s %(funcName)s %(levelname)s %(message)s', level=logging.INFO)
logger = logging.getLogger("bitfinex-basic-example")

env_path = os.path.join(os.path.dirname(__file__), '.env')
load_dotenv(env_path)

auth_param = dict(apiKey=os.environ.get('CEX_API_KEY'),
                  secret=os.environ.get('CEX_SECRET'),
                  uid=os.environ.get('CEX_UID'))

print(auth_param)

conn = Cex(auth_param)
parser = CexParser()
watcher = BookWatcher(parser)
conn.on_message = parser.parse


async def subscribe():
    for market in ["BTC-EUR", 'ETH-EUR', "ETH-BTC"]:
        pair = list(market.split('-'))
        request = {'e': "order-book-subscribe",
                   'oid': 'book-1',
                   'data': {'pair': [pair[0], pair[1]], 'subscribe': True, 'depth': 0}
                   }

        await conn.ws.send_json(request)


output = open("cex_dump_1h.jsonl", "w")


def dump(msg: str):
    output.write(msg)
    output.write('\n')


conn.on_connect_ws = subscribe
conn.on_before_handling = dump
loop = asyncio.get_event_loop()


def stop():
    asyncio.ensure_future(conn.ws.close())


loop.call_later(3600, stop)
try:
    loop.run_until_complete(conn.run_receiver())
except KeyboardInterrupt:
    print("Application closed by KeyboardInterrupt.")



from sortedcontainers import SortedKeyList

VERY_SMALL_NUMBER = 1e-11


class L2BookSide:

    def __init__(self, is_bids: bool):
        if is_bids:
            self.data = SortedKeyList(key=lambda val: -val[0])
        else:
            self.data = SortedKeyList(key=lambda val: val[0])
        self.is_bids = is_bids
        self.time = None

    def fill(self, source):
        self.data.clear()
        for item in source:
            self.add(item)

    def add(self, price: float, size: float):
        self.data.add([price, size])

    def update(self, price: float, size: float):
        key = -price if self.is_bids else price
        i = self.data.bisect_key_left(key)

        if 0 <= i < len(self.data):
            value = self.data[i]
        else:
            if size <= VERY_SMALL_NUMBER:
                return False

            self.data.add([price, size])
            return True

        if size <= VERY_SMALL_NUMBER:
            if value[0] == price:
                old_size = self.data[i][1]
                self.data.discard(value)
                return True
            else:
                return False

        if value[0] == price:
            self.data[i][1] = size
        else:
            self.data.add([price, size])
        return True

    def delete(self, price: float):
        return self.update(price, 0.0)



import asyncio
import json
import logging
from typing import Callable, Any, Coroutine, Awaitable, Optional

import aiohttp
from time import time

logger = logging.getLogger(__name__)


class SessionWrapper:
    """

    Handlers:
     - on_connect_ws - Called after the connection is established.
            If auth_type is not equivalent to AuthType.NONE,
            on_connect_ws will be set to self.auth_login;
     - on_close_ws - Called after disconnection, default value is None;
     - on_authenticated - Called after authentication is confirmed, default value is None;
     - on_message - Called when a message is received, default value is self.handle_message;
     - on_before_handling -
    """

    def __init__(self, timeout: aiohttp.ClientTimeout = None, session: aiohttp.ClientSession = None):
        self.__internal_session: bool = True
        self.__is_session: bool = False
        self.__session: Optional[aiohttp.ClientSession] = None
        self._timeout: Optional[aiohttp.ClientTimeout] = None

        self._on_connect: Optional[Callable[[], Any]] = None
        self._async_on_connect: Optional[Callable[[], Awaitable[Any]]] = None

        self._on_message: Optional[Callable[[str], bool]] = None
        self._async_on_message: Optional[Callable[[str], Awaitable[bool]]] = None

        self.on_close_ws: Optional[Callable[[], Any]] = None
        self.on_before_handling: Optional[Callable[[str], Any]] = None

        self.ws_api: str = ""
        self.rest_api: str = ""
        self.ws: Optional[aiohttp.ClientWebSocketResponse] = None
        self.last_message: str = ""
        self.receipt_time: float = 0.

        if session:
            self._session = session
            self.__internal_session = False

        self._timeout = timeout if timeout else aiohttp.ClientTimeout(total=20)

        self.logger = logging.getLogger(__name__)

    @property
    def _session(self):
        if not self.__is_session:
            self.__check_session()
        return self.__session

    @_session.setter
    def _session(self, value):
        self.__is_session = True
        self.__internal_session = False
        self.__session = value

    def __check_session(self):
        if not self.__is_session:
            self.__is_session = True
            self.__internal_session = True
            self.__session = aiohttp.ClientSession(timeout=self._timeout)

    def __del__(self):
        self._close()

    def __enter__(self):
        self.__check_session()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._close()

    def _close(self):
        if self.__internal_session and self.__session:
            asyncio.ensure_future(self.__session.close())
            self.__is_session = False

    @property
    def on_message(self):
        return self._on_message

    @on_message.setter
    def on_message(self, handler):
        if handler is None:
            self._async_on_message = None
            self._on_message = None
        else:
            if asyncio.iscoroutinefunction(handler):
                self._async_on_message = handler
            else:
                self._on_message = handler

    @property
    def on_connect_ws(self):
        if self._async_on_connect:
            return self._async_on_connect
        else:
            return self._on_connect

    @on_connect_ws.setter
    def on_connect_ws(self, handler):
        if handler is None:
            self._async_on_connect = None
            self._on_connect = None
        else:
            if asyncio.iscoroutinefunction(handler):
                self._async_on_connect = handler
                self._on_connect = None
            else:
                self._async_on_connect = None
                self._on_connect = handler

    async def run_receiver(self):
        """
        Establish a connection and start the receiver loop.
        :return:
        """
        self.ws = await self._session.ws_connect(self.ws_api)
        if self._async_on_connect:
            await self._async_on_connect()
        elif self._on_connect:
            self._on_connect()

        # A receiver loop
        while self.ws and not self.ws.closed:
            message = await self.ws.receive()
            self.receipt_time = time()

            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.ERROR):
                self.logger.warning(f"Connection close {repr(message)}")
                if self.on_close_ws:
                    await self.on_close_ws()

                continue
            if message.type == aiohttp.WSMsgType.CLOSING:
                self.logger.debug(f"Connection closing {repr(message)}")
                continue

            if message.type == aiohttp.WSMsgType.TEXT:
                self.last_message = message.data
                if self.on_before_handling:
                    self.on_before_handling(message.data)

                processed = False
                if self._async_on_message:
                    processed = await self._async_on_message(message.data)
                elif self._on_message:
                    processed = self._on_message(message.data)

                if not processed:
                    self.handle_message(message.data)
            else:
                self.logger.warning(f"Unknown type of message {repr(message)}")

    def handle_message(self, message: str):
        pass

    async def public_get(self, request_path, params=None):
        async with self._session.get(url=self.rest_api + request_path,
                                     params=params,
                                     headers=None) as response:
            if response.status == 200:
                data = await response.read()
                if response.content_type.endswith('json'):
                    data = json.loads(data)
                else:
                    logger.warning(f"Unexpected response.content_type:{response.content_type}")
                return data
            else:
                response.raise_for_status()

    async def stop(self):
        await self.ws.close()
        await self._session.close()



from .l2_book_side import L2BookSide


class L2Book:
    """

    """

    def __init__(self, instrument: str):
        """

        :param instrument:
        """
        self._instrument = instrument
        self._bids = L2BookSide(is_bids=True)
        self._asks = L2BookSide(is_bids=False)

    def instrument(self) -> str:
        return self._instrument

    @property
    def bids(self) -> L2BookSide:
        return self._bids

    @property
    def asks(self) -> L2BookSide:
        return self._asks

    def clear(self):
        self._bids.data.clear()
        self._asks.data.clear()

    def top_ask_price(self):
        return self._asks.data[0][0]

    def top_bid_price(self):
        return self._bids.data[0][0]

    def valid(self) -> bool:
        return len(self._asks.data) != 0 and len(self._bids.data) != 0



import asyncio
import json
import logging
from time import time
from typing import Optional, Callable, Any, Awaitable

import aiohttp

from ssc2ce.common.session import SessionWrapper
from ssc2ce.common.utils import resolve_route

from enum import IntEnum


class Bitfinex(SessionWrapper):
    _user_on_connect: Optional[Callable[[], Any]] = None
    _user_async_on_connect: Optional[Callable[[], Awaitable[Any]]] = None

    class ConfigFlag(IntEnum):
        TIMESTAMP = 32768
        SEQ_ALL = 65536
        CHECKSUM = 131072

    class StatusFlag(IntEnum):
        MAINTENANCE = 0
        OPERATIVE = 1

    on_maintenance = None
    on_conf = None
    receipt_time = None
    is_connected = False
    subscriptions = []
    channel_handlers = {}

    def __init__(self,
                 flags: ConfigFlag = ConfigFlag.TIMESTAMP | ConfigFlag.SEQ_ALL):
        super().__init__()

        self.ws_api = 'wss://api-pub.bitfinex.com/ws/2'
        self.flags = flags
        self.logger = logging.getLogger(__name__)
        self.routes = [
            ("subscribed", self.handle_subscribed),
            ("info", self.handle_info),
            ("conf", self.handle_conf),
            # ("", self.warn_handler),
        ]

    async def configure(self, flags: ConfigFlag = None):
        if flags is not None:
            self.flags = flags

        if self.flags is not None:
            request = dict(event="conf", flags=self.flags)
            await self.ws.send_json(request)

    async def subscribe(self, request, handler):
        self.subscriptions.append((request, handler))
        self.logger.info(f"subscribe {request}")
        await self.ws.send_json({
            "event": "subscribe",
            **request
        })

    @property
    def on_connect_ws(self):
        if self._async_on_connect:
            return self._user_async_on_connect
        else:
            return self._user_on_connect

    @on_connect_ws.setter
    def on_connect_ws(self, handler):
        if handler is None:
            self._user_async_on_connect = None
            self._user_on_connect = None
        else:
            if asyncio.iscoroutinefunction(handler):
                self._user_async_on_connect = handler
                self._user_on_connect = None
            else:
                self._user_async_on_connect = None
                self._user_on_connect = handler

    def handle_message(self, message: str):
        data = json.loads(message)

        if isinstance(data, list):
            channel_id = data[0]
            handler = self.channel_handlers.get(channel_id)
            if handler:
                handler(data, self)
            else:
                self.logger.warning(f"Can't find handler for channel_id{channel_id}, {message}")
        elif isinstance(data, dict):
            if "event" in data:
                if asyncio.iscoroutinefunction(self.handle_event):
                    asyncio.ensure_future(self.handle_event(data))
                else:
                    self.handle_event(data)
            else:
                self.logger.warning(f"Unknown message {message}")
        else:
            self.logger.warning(f"Unknown message {message}")

    async def empty_handler(self, data):
        pass

    def handle_subscribed(self, message):
        self.logger.info(f"receive subscribed: {message}")
        idx = None
        for i, s in enumerate(self.subscriptions):
            if s[0].items() <= message.items():
                idx = i
                self.channel_handlers[message["chanId"]] = s[1]

        if idx:
            del self.subscriptions[idx]

    def handle_conf(self, message):
        """{'event': 'conf', 'status': 'OK', 'flags': 98304}"""
        self.logger.info(f"{message}")
        if self.on_conf:
            self.on_conf()

    def handle_info(self, message):
        """
        Handle an info message that contains the actual version of the websocket stream, along with a platform status
        flag (1 for operative, 0 for maintenance
        :param message: Info messages {'event': 'info', 'version': 2, 'serverId': ..., 'platform': {'status': 1}}
          version: the actual version of the websocket stream must be 2
          status: a platform status flag (1 for operative, 0 for maintenance).
        :return:
        """
        self.logger.info(f"{message}")

        if message["version"] != 2:
            raise NotImplemented(f"Bitfinex connector support only version 2 but receive {message}")

        if message["platform"]["status"] == 1:
            if not self.is_connected:
                asyncio.ensure_future(self.configure())
                if self._user_async_on_connect:
                    asyncio.ensure_future(self._user_async_on_connect())
                elif self._user_on_connect:
                    self._user_on_connect()
            else:
                if self.on_maintenance:
                    asyncio.ensure_future(self.on_maintenance(message))
        else:
            if self.on_maintenance:
                asyncio.ensure_future(self.on_maintenance(message))

    def warn_handler(self, message):
        self.logger.warning(f"Unsupported message {message}")

    def handle_event(self, message):
        event = message["event"]
        handler = resolve_route(event, self.routes)

        if handler:
            handler(message)
        else:
            self.logger.warning(f"Unhandled event:{event}, message:{repr(message)} .")
        return

    def close(self):
        super()._close()

    async def stop(self):
        await self.ws.close()
        self.close()



from .deribit import Deribit, DeribitParser, DeribitL2Book, AuthType
from .bitfinex import Bitfinex, BitfinexL2Book
from .coinbase_pro import Coinbase, CoinbaseParser, CoinbaseL2Book
from .cex import Cex, CexParser, CexL2Book
from .factory import create_parser
from .common.exceptions import BrokenOrderbook
from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



import logging
from ssc2ce.common import L2Book


class BitfinexL2Book(L2Book):
    change_id = None
    timestamp = None
    logger = logging.getLogger(__name__)

    def __init__(self, instrument: str):
        """

        :param instrument:
        """
        L2Book.__init__(self, instrument)

    def handle_snapshot(self, message: dict) -> None:
        """

        :param message:
        :return:
        """
        self.clear()

        for item in message[1]:
            price = item[2]
            if price < 0:
                self.asks.add(item[0], -price)
            else:
                self.bids.add(item[0], price)

    def handle_update(self, message: dict) -> None:
        """

        :param message:
        :return:
        """
        price = message[1][0]
        size = message[1][2]
        count = message[1][1]
        if size < 0:
            if count == 0:
                self._asks.update(price, 0.)
            else:
                self._asks.update(price, -size)
        else:
            if count == 0:
                self._bids.update(price, 0.)
            else:
                self._bids.update(price, size)



class Ssc2ceError(Exception):
    """Base class for errors."""


class BrokenOrderbook(Exception):
    def __init__(self, instrument, prev_change_id, change_id):
        self.instrument = instrument
        self.message = f"instrument:{self.instrument} expected:{change_id} != received:{prev_change_id}"



import os

from unittest import TestCase

from ssc2ce import CexParser, BrokenOrderbook
from .cex_expected_result import results

path = os.path.dirname(os.path.abspath(__file__))
path = os.path.join(path, "cex_dump.jsonl")
with open(path) as f:
    lines = f.readlines()


class TestCexParser(TestCase):
    broken = False

    def test_one(self):
        parser = CexParser()
        book = parser.get_book("BTC:EUR")
        for i, line in enumerate(lines):
            # print(results[i])
            if results[i][0] == 1:
                parser.parse(line)
                self.assertTrue(book.valid())
                self.assertEqual(book.top_bid_price(), results[i][1])
                self.assertEqual(book.top_ask_price(), results[i][2])
            elif results[i][0] == 0:
                parser.parse(line)
                self.assertFalse(book.valid())
            else:
                self.assertRaises(BrokenOrderbook, parser.parse, line)

    def handle_broken(self, pair: str, expected: int, received: int):
        print(pair, expected, received)
        self.broken = True

    def test_two(self):
        parser = CexParser()
        parser._handler_broken_orderbook = self.handle_broken

        book = parser.get_book("BTC:EUR")
        for i, line in enumerate(lines):
            # print(results[i])
            if results[i][0] == 1:
                parser.parse(line)
                self.assertTrue(book.valid())
                self.assertEqual(book.top_bid_price(), results[i][1])
                self.assertEqual(book.top_ask_price(), results[i][2])
                self.assertFalse(self.broken)
            elif results[i][0] == 0:
                parser.parse(line)
                self.assertFalse(book.valid())
                self.assertFalse(self.broken)
            else:
                parser.parse(line)
                self.assertTrue(self.broken)


import logging

logger = logging.getLogger(__name__)


def create_parser(exchange: str, is_cpp: bool):
    if is_cpp:
        from importlib import util

        ssc2ce_cpp_spec = util.find_spec("ssc2ce_cpp")
        if ssc2ce_cpp_spec is None:
            logger.error("You must install the ssc2ce_cpp module to use its features.\n pip install ssc2ce_cpp")
            return None

        if exchange == 'deribit':
            from ssc2ce_cpp import DeribitParser as Parser
        elif exchange == 'coinbase':
            from ssc2ce_cpp import CoinbaseParser as Parser
        elif exchange == 'cex':
            from ssc2ce_cpp import CexParser as Parser
        else:
            logger.error(f"Unknown exchange {exchange}")
            return None
    else:
        if exchange == 'deribit':
            from ssc2ce.deribit.parser import DeribitParser as Parser
        elif exchange == 'coinbase':
            from ssc2ce.coinbase_pro.parser import CoinbaseParser as Parser
        elif exchange == 'cex':
            from ssc2ce.cex.parser import CexParser as Parser
        else:
            logger.error(f"Unknown exchange {exchange}")
            return None

    return Parser()



import asyncio
import json
import logging
from abc import ABC, abstractmethod
from typing import Callable, Optional

from ssc2ce.deribit.l2_book import DeribitL2Book, L2Book
from ssc2ce.common.exceptions import BrokenOrderbook


class AbstractParser(ABC):
    """

    """

    def __init__(self):
        self._on_book_update: Optional[Callable[[L2Book], None]] = None
        self._on_book_setup: Optional[Callable[[L2Book], None]] = None

    def set_on_book_setup(self, handler) -> None:
        self._on_book_setup = handler

    def set_on_book_update(self, handler) -> None:
        self._on_book_update = handler

    @abstractmethod
    def parse(self, message: str) -> bool:
        return False



from ssc2ce.common.abstract_parser import AbstractParser
from ssc2ce.deribit.l2_book import L2Book


class BookWatcher:
    def __init__(self, parser: AbstractParser, print_it: bool = True):
        self.betters = {}
        self.print_it = print_it
        parser.set_on_book_setup(self.handle_book_setup)
        parser.set_on_book_update(self.handle_book_update)

    def handle_book_setup(self, book: L2Book):
        top = [book.top_bid_price(), book.top_ask_price()]
        self.betters[book.instrument()] = top
        if self.print_it:
            print(f"{book.instrument()} bid:{top[0]} ask:{top[1]}")

    def handle_book_update(self, book: L2Book):
        top = self.betters[book.instrument()]
        if top[1] != book.top_ask_price() or top[0] != book.top_bid_price():
            top[1] = book.top_ask_price()
            top[0] = book.top_bid_price()
            if self.print_it:
                print(f"{book.instrument()} bid:{top[0]} ask:{top[1]}")



import asyncio
import json
import logging

import aiohttp

from ssc2ce.common.exceptions import Ssc2ceError
from ssc2ce.common.session import SessionWrapper
from ssc2ce.common.utils import resolve_route, IntId

from enum import IntEnum


class AuthType(IntEnum):
    NONE = 0
    PASSWORD = 1
    CREDENTIALS = 2
    SIGNATURE = 3


def hide_secret(request):
    data = request.copy()
    hidden_keys = ["password", "username", "client_id", "client_secret", "refresh_token", "access_token"]

    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = hide_secret(value)
        elif key in hidden_keys:
            data[key] = "***"

    return data


class Deribit(SessionWrapper):
    """

    Handlers:
     - on_connect_ws - Called after the connection is established.
            If auth_type is not equivalent to AuthType.NONE,
            on_connect_ws will be set to self.auth_login;
     - on_close_ws - Called after disconnection, default value is None;
     - on_authenticated - Called after authentication is confirmed, default value is None;
     - on_token - Called after receiving a response to an authentication request, default value is None;
     - on_message - Called when a message is received, default value is self.handle_message;
     - on_handle_response - Called when the message from the exchange does not contain the request id;
     - on_response_error - Called when the response contains an error message.
    """
    auth_params: dict = None

    def __init__(self,
                 client_id: str = None,
                 client_secret: str = None,
                 scope: str = "session",
                 testnet: bool = True,
                 auth_type: AuthType = AuthType.NONE,
                 get_id=IntId().get_id):

        if auth_type & (AuthType.CREDENTIALS | AuthType.SIGNATURE):
            if client_secret is None or client_id is None:
                raise Ssc2ceError(f" Authentication {str(auth_type)} need client_id and client_secret")

        if auth_type == AuthType.SIGNATURE:
            raise NotImplemented(f"Authentication {str(auth_type)} for Deribit is not implemented.")

        super().__init__()

        if auth_type != AuthType.NONE:
            self.on_connect_ws = self.auth_login

        self.on_authenticated = None
        self.on_token = None
        self.on_response_error = None
        self.on_handle_response = None

        self.requests = {}
        self.ws_api = f"wss://{'test' if testnet else 'www'}.deribit.com/ws/api/v2/"
        self.get_id = get_id
        self.logger = logging.getLogger(__name__)
        self.auth_type = auth_type
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope

        self.method_routes = [
            ("heartbeat", self.handle_heartbeat),
        ]
        self.response_routes = [
            ("public/auth", self.handle_auth),
            ("", self.empty_handler),
        ]

    def close(self):
        super()._close()

    async def stop(self):
        """
        Close connection and break the receiver loop
        :return:
        """
        await self.ws.close()

    async def send_public(self, request: dict, callback=None, logging_it: bool = True) -> int:
        """
        Send a public request

        :param request: Request without jsonrpc and id fields
        :param callback: The function that will be called after receiving the query result. Default is None
        :param logging_it:
        :return: Request Id
        """
        request_id = self.get_id()
        request = {
            "jsonrpc": "2.0",
            "id": request_id,
            **request
        }
        if logging_it:
            self.logger.info(f"sending:{repr(hide_secret(request))}")
        await self.ws.send_json(request)

        if callback:
            request["callback"] = callback

        self.requests[request_id] = request

        return request_id

    async def send_private(self, request: dict, callback=None) -> int:
        """
        Send a private request

        :param request: Request without jsonrpc and id fields
        :param callback: The function that will be called after receiving the query result. Default is None
        :return: Request Id
        """

        request_id = self.get_id()
        access_token = self.auth_params["access_token"]
        request["params"]["access_token"] = access_token

        request = {
            "jsonrpc": "2.0",
            "id": request_id,
            **request
        }
        self.logger.info(f"sending:{repr(hide_secret(request))}")
        await self.ws.send_json(request)

        if callback:
            request["callback"] = callback

        self.requests[request_id] = request

        return request_id

    async def send(self, request: dict, callback=None) -> int:
        """
        A wrapper for send_private and send_public, defines the type of request by content.

        :param request: Request without jsonrpc and id fields
        :param callback: The function that will be called after receiving the query result. Default is None
        :return: Request Id
        """
        method = request["method"]
        if method.starts("public/"):
            request_id = await self.send_public(request, callback)
        else:
            request_id = await self.send_private(request, callback)

        return request_id

    async def auth_login(self) -> int:
        """
        Send an authentication request using parameters stored in the constructor

        :return: Reuest id
        """
        self.auth_type = AuthType.CREDENTIALS

        msg = {
            "method": "public/auth",
            "params": {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }
        }

        if self.scope:
            msg["scope"] = self.scope

        request_id = await self.send_public(msg)
        return request_id

    async def auth_client_credentials(self, client_id, client_secret, scope: str = None) -> int:
        """
        Send a credentials authentication request

        :param client_id: using the access key
        :param client_secret: and access secret that can be found on the API page on the website
        :param scope:
            connection, session, session:name,
            trade:[read, read_write, none],
            wallet:[read, read_write, none],
            account:[read, read_write, none]
        :return: Request Id
        """

        self.auth_type = AuthType.CREDENTIALS

        msg = {
            "method": "public/auth",
            "params": {
                "grant_type": "client_credentials",
                "client_id": client_id,
                "client_secret": client_secret,
            }
        }

        if scope:
            self.scope = scope

        if self.scope:
            msg["scope"] = self.scope

        request_id = await self.send_public(msg)
        return request_id

    async def auth_password(self, username: str, password: str, scope: str = None) -> int:
        """
        Send a password authentication request

        :param username: User name
        :param password: Password
        :param scope:
            connection, session, session:name,
            trade:[read, read_write, none],
            wallet:[read, read_write, none],
            account:[read, read_write, none]
        :return: Request Id
        """

        self.auth_type = AuthType.PASSWORD
        msg = {
            "method": "public/auth",
            "params": {
                "grant_type": "password",
                "username": username,
                "password": password
            }
        }

        if scope:
            self.scope = scope

        if self.scope:
            msg["scope"] = scope

        request_id = await self.send_public(msg)
        return request_id

    async def auth_refresh_token(self) -> int:
        """
        Send a refresh token request

        :return: Request Id
        """
        msg = {
            "method": "public/auth",
            "params": {
                "grant_type": "refresh_token",
                "refresh_token": self.auth_params["refresh_token"]
            }
        }

        request_id = await self.send_public(msg)
        return request_id

    async def auth_logout(self) -> int:
        """
        Send a logout request

        :return: Request Id
        """
        msg = {
            "method": "private/logout",
            "params": {}
        }

        request_id = await self.send_private(msg)
        return request_id

    async def set_heartbeat(self, interval: int = 15) -> int:
        """

        :param interval:
        :return:
        """
        request_id = await self.send_public({"method": "public/set_heartbeat", "params": {"interval": interval}})
        return request_id

    async def disable_heartbeat(self) -> int:
        """

        :return:
        """
        request_id = await self.send_public({"method": "public/disable_heartbeat", "params": {}})
        return request_id

    async def get_currencies(self, callback=None):
        """
        Send a request for a list of cryptocurrencies supported by the API
        :param callback:
        :return: Request Id
        """
        return await self.send_public(request=dict(method="public/get_currencies", params={}), callback=callback)

    async def get_instruments(self, currency: str, kind: str = None, expired: bool = False, callback=None) -> int:
        """
        Send a request for a list available trading instruments
        :param currency: The currency symbol: BTC or ETH
        :param kind: Instrument kind: future or option, if not provided instruments of all kinds are considered
        :param expired:
        :param callback:
        :return: Request Id
        """
        request = {"method": "public/get_instruments",
                   "params": {
                       "currency": currency,
                       "expired": expired
                   }}
        if kind:
            request["params"]["kind"] = kind

        return await self.send_public(request=request, callback=callback)

    def handle_message(self, message: str) -> None:
        """

        :param message:
        :return:
        """
        data = json.loads(message)
        if "method" in data:
            self.handle_method_message(data)
        else:
            if "id" in data:
                if "error" in data:
                    self.handle_error(data)
                else:
                    request_id = data["id"]
                    if request_id:
                        self.handle_response(request_id, data)
            else:
                self.handle_method_message(data)

    def empty_handler(self, **kwargs) -> None:
        """
        A default handler
        :param kwargs:
        :return:
        """
        self.logger.debug(f"{repr(kwargs)}")

    def handle_response(self, request_id: int, response: dict) -> None:
        """

        :param request_id:
        :param response:
        :return:
        """

        request = self.requests.get(request_id)
        if request:
            if "callback" in request:
                callback = request["callback"]
                if asyncio.iscoroutinefunction(callback):
                    asyncio.ensure_future(callback(response))
                else:
                    callback(response)
            else:
                method = request["method"]
                handler = resolve_route(method, self.response_routes)
                if handler:
                    if asyncio.iscoroutinefunction(handler):
                        asyncio.ensure_future(handler(request=request, response=response))
                    else:
                        handler(request=request, response=response)
                else:
                    self.logger.warning(f"Unhandled method:{method} response:{repr(response)} "
                                        f"to request:{repr(request)}.")

            del self.requests[request_id]
        else:
            if self.on_handle_response:
                if asyncio.iscoroutinefunction(self.on_handle_response):
                    asyncio.ensure_future(self.on_handle_response(response))
                else:
                    self.on_handle_response(response)
                return
            else:
                self.logger.warning(
                    f"Unknown id:{request_id}, the on_handle_response event must be defined."
                    f" Unhandled message {response}")

    def handle_method_message(self, data) -> None:
        """

        :param data:
        :return:
        """
        method = data["method"]
        handler = resolve_route(method, self.method_routes)

        if handler:
            if asyncio.iscoroutinefunction(handler):
                asyncio.ensure_future(handler(data))
            else:
                handler(data)

        elif not self.on_before_handling:
            self.logger.warning(f"Unhandled message:{repr(data)}.")

    def handle_heartbeat(self, data):
        """

        :param data:
        :return:
        """
        if data["params"]["type"] == "test_request":
            asyncio.ensure_future(self.send_public({"method": "public/test", "params": {}}, logging_it=False))

    def handle_auth(self, request, response) -> None:
        """

        :param request:
        :param response:
        :return:
        """
        self.auth_params = response["result"]
        grant_type = request["params"]["grant_type"]
        if grant_type == "":
            if self.on_token:
                self.on_token(response["result"])
        elif grant_type in ("client_credentials", "password"):
            # TODO - Why we use two handlers?
            if self.on_authenticated:
                self.on_authenticated()
            if self.on_token:
                self.on_token(response["result"])
        elif grant_type == "client_signature":
            pass
        else:
            self.logger.error(f"Unknown grant_type {repr(hide_secret(request))} : {repr(hide_secret(response))}")

    def handle_error(self, message):
        if self.on_response_error:
            self.on_response_error(message)
        else:
            self.logger.error(message)



import asyncio
import json
import logging
import hmac
import hashlib
from datetime import datetime, timezone
from typing import Optional, Callable, Any, Awaitable

import aiohttp
from aiohttp import ClientResponseError
from ssc2ce.common.session import SessionWrapper


class WrongMessage(Exception):
    def __init__(self, msg: aiohttp.WSMessage):
        self.msg = msg


def convert_message(msg: aiohttp.WSMessage):
    if msg.type == aiohttp.WSMsgType.TEXT:
        message = msg.data
        return json.loads(message)
    else:
        raise WrongMessage(msg)


def hide_secret(request: dict):
    data = request.copy()
    hidden_keys = ["key", "signature"]

    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = hide_secret(value)
        elif key in hidden_keys:
            data[key] = "***"

    return data


class Cex(SessionWrapper):
    """

    Handlers:
     - on_connect_ws - Called after the connection is established.
            If auth_type is not equivalent to AuthType.NONE,
            on_connect_ws will be set to self.auth_login;
     - on_close_ws - Called after disconnection, default value is None;
     - on_authenticated - Called after authentication is confirmed, default value is None;
     - on_token - Called after receiving a response to an authentication request, default value is None;
     - on_message - Called when a message is received, default value is self.handle_message;
     - on_handle_response - Called when the message from the exchange does not contain the request id;
     - on_response_error - Called when the response contains an error message.
    """

    _user_on_connect: Optional[Callable[[], Any]] = None
    _user_async_on_connect: Optional[Callable[[], Awaitable[Any]]] = None

    def __init__(self, auth_param: dict = None):

        super().__init__()

        self.ws_api = 'wss://ws.cex.io/ws/'
        self.rest_api = 'https://cex.io/api/'
        self.logger = logging.getLogger(__name__)
        self.auth_param = auth_param.copy() if auth_param else auth_param
        self._async_on_connect = self.handle_connected

    def rest_signature(self):
        """
        Create signature for REST request

        Look at https://cex.io/rest-api#/definitions/signature
        An HMAC-SHA256 encoded message containing - a nonce, user ID and API key. The HMAC-SHA256 code must be
        enerated using a secret key that was generated with your API key. This code must be converted to
        its hexadecimal representation (64 uppercase characters).

        :return:
        """

        unix_ts = int(datetime.now().timestamp())  # UNIX timestamp in seconds
        string = "{}{}{}".format(unix_ts, self.auth_param["uid"], self.auth_param['apiKey'])
        signature = hmac.new(self.auth_param['secret'].encode(), string.encode(), hashlib.sha256).hexdigest().upper()
        return {'key': self.auth_param['apiKey'],
                'signature': signature,
                'nonce': unix_ts, }

    async def authorize(self):
        """

        :return:
        """

        unix_ts = int(datetime.now().timestamp())  # UNIX timestamp in seconds
        string = "{}{}".format(unix_ts, self.auth_param['apiKey'])
        signature = hmac.new(self.auth_param['secret'].encode(), string.encode(), hashlib.sha256).hexdigest()
        request = {'e': 'auth',
                   'auth': {'key': self.auth_param["apiKey"],
                            'signature': signature,
                            'timestamp': unix_ts},
                   'oid': 'auth', }

        try:
            self.logger.info(f"Auth{hide_secret(request)} is been sending")
            await self.ws.send_json(request)
        except Exception as e:
            logging.error("{}.authorize, {}({})".format(self.__class__.__name__, e.__class__.__name__, e))
            raise

        msg = await self.ws.receive()
        msg = convert_message(msg)
        if msg.get('e') == 'auth' and msg.get('ok') == 'ok' and msg["data"].get('ok') == 'ok':
            conf_ts = msg["timestamp"]
            conf_ts = datetime.fromtimestamp(conf_ts, tz=timezone.utc).isoformat()
            self.logger.info("Successful authorization. " + conf_ts)
        else:
            self.logger.error("Unsuccessful authorization:" + json.dumps(msg))
            await self.ws.close()

    @property
    def on_connect_ws(self):
        if self._async_on_connect:
            return self._user_async_on_connect
        else:
            return self._user_on_connect

    @on_connect_ws.setter
    def on_connect_ws(self, handler):
        if handler is None:
            self._user_async_on_connect = None
            self._user_on_connect = None
        else:
            if asyncio.iscoroutinefunction(handler):
                self._user_async_on_connect = handler
                self._user_on_connect = None
            else:
                self._user_async_on_connect = None
                self._user_on_connect = handler

    async def handle_connected(self):
        msg = await self.ws.receive()
        msg = convert_message(msg)
        if msg.get('e') == 'connected':
            if self.auth_param:
                await self.authorize()

            if self._user_async_on_connect:
                await self._user_async_on_connect()
            elif self._user_on_connect:
                self._user_on_connect()
        else:
            self.logger.error(f"Wrong protocol. Expected 'connected' received:{json.dumps(msg)}")
            await self.ws.close()

    async def ping(self):
        request = '{"e": "pong"}'
        await self.ws.send_str(request)
        self.logger.debug("ping - " + json.dumps(request))

    def handle_message(self, message: str):
        data = json.loads(message)

        msg_type = data.get('e')
        if msg_type == 'ping':
            asyncio.ensure_future(self.ping())
        else:
            self.logger.info(message)

    async def get_currency_limits(self):
        try:
            result = await self.public_get("currency_limits")
            return result
        except ClientResponseError as e:
            self.logger.error(e.message)
            return {}

    async def private_get(self, request_path, params: dict = None):
        if params is None:
            params = self.rest_signature()
        else:
            params.update(self.rest_signature())

        print(params)
        # body = json.dumps(params)

        async with self._session.get(self.rest_api + request_path, data=params) as response:
            if response.status == 200:
                data = await response.read()
                data = json.loads(data)
                return data
            else:
                self.logger.error(self.rest_api + request_path)
                response.raise_for_status()



import json
import logging
from typing import Callable

from ssc2ce.coinbase_pro.l2_book import L2Book
from ssc2ce.common.abstract_parser import AbstractParser


class CoinbaseParser(AbstractParser):
    """

    """

    _on_subscriptions: Callable[[dict], bool] = None
    _on_heartbeat: Callable[[dict], bool] = None

    def __init__(self):
        AbstractParser.__init__(self)
        self.logger = logging.getLogger(__name__)
        self.deprecated_already_warn = False
        self.books = {}
        self.routes = {
            "l2update": self.handle_update,
            "snapshot": self.handle_snapshot,
            "subscriptions": self.handle_subscriptions,
            "heartbeat": self.handle_heartbeat
        }

    def set_on_subscriptions(self, handler) -> None:
        self._on_subscriptions = handler

    def set_on_heartbeat(self, handler) -> None:
        self._on_heartbeat = handler

    def parse(self, message) -> bool:
        data = json.loads(message)

        if isinstance(data, dict):
            handler: Callable[[dict], bool] = self.routes.get(data.get('type'))
            if handler:
                return handler(data)

            return False

    def get_book(self, instrument: str) -> L2Book:
        book: L2Book = self.books.get(instrument)
        if book is None:
            book = L2Book(instrument)
            self.books[instrument] = book

        return book

    def handle_snapshot(self, message: dict) -> bool:
        """

        :param message:
        :return:
        """
        book = self.get_book(message["product_id"])
        book.clear()

        for i in message['bids']:
            book.bids.add(float(i[0]), float(i[1]))

        for i in message['asks']:
            book.asks.add(float(i[0]), float(i[1]))

        if self._on_book_setup:
            self._on_book_setup(book)

        return True

    def handle_update(self, message: dict) -> bool:
        """

        :param message:
        :return:
        """
        book = self.get_book(message["product_id"])
        book.time = message["time"]
        for change in message['changes']:
            if change[0] == 'sell':
                book.asks.update(price=float(change[1]), size=float(change[2]))
            else:
                book.bids.update(price=float(change[1]), size=float(change[2]))

        if self._on_book_update:
            self._on_book_update(book)
        return True

    def handle_subscriptions(self, data: dict) -> bool:
        if self._on_subscriptions:
            return self._on_subscriptions(data)

        return False

    def handle_heartbeat(self, data: dict) -> bool:
        if self._on_heartbeat:
            return self._on_heartbeat(data)

        return False



from ssc2ce.common.l2_book import L2Book


class CoinbaseL2Book(L2Book):
    def __init__(self, instrument: str):
        L2Book.__init__(self, instrument)
        self.change_id = None
        self.timestamp = None



#!/usr/bin/env python
import asyncio
import logging
from ssc2ce import Coinbase

conn = Coinbase(sandbox=False)

pending = {}

logging.basicConfig(format='%(asctime)s %(name)s %(funcName)s %(levelname)s %(message)s', level=logging.INFO)
logger = logging.getLogger("deribit-writer")

instruments = []


async def subscribe_books():
    await conn.ws.send_json({
        "type": "subscribe",
        "product_ids": instruments,
        "channels": [
            "level2",
            "heartbeat"
        ]
    })


output = open("coinbase_dump.jsonl", "w")


def dump(msg: str):
    output.write(msg)
    output.write('\n')


def stop():
    asyncio.ensure_future(conn.ws.close())


async def run():
    all_products = [x for x in await conn.get_products()]

    my_products = {x['id']: x for x in all_products if x['quote_currency'] == 'EUR'}
    base_currencies = [x['base_currency'] for x in my_products.values()]

    for product in all_products:
        if product["id"] in my_products:
            continue

        if product['quote_currency'] in base_currencies:
            my_products[product["id"]] = product

    global instruments
    instruments += list(my_products.keys())
    await conn.run_receiver()


conn.on_before_handling = dump
conn.on_connect_ws = subscribe_books

loop = asyncio.get_event_loop()
loop.call_later(3600, stop)

try:
    loop.run_until_complete(run())
except KeyboardInterrupt:
    print("Application closed by KeyboardInterrupt.")
